"""Testing SBML functionality based on libsbml."""

from collections import namedtuple
from os import unlink
from os.path import join, split
from pathlib import Path
from pickle import load
from tempfile import gettempdir
from typing import List, Tuple

import pytest
from _pytest.fixtures import SubRequest

import cobra
from cobra import Model
from cobra.io import read_sbml_model, validate_sbml_model, write_sbml_model


config = cobra.Configuration()  # for default bounds

try:
    import jsonschema
except ImportError:
    jsonschema = None

# ----------------------------------
# Definition of SBML files to test
# ----------------------------------
IOTrial = namedtuple(
    "IOTrial",
    [
        "name",
        "reference_file",
        "test_file",
        "read_function",
        "write_function",
        "validation_function",
    ],
)
trials: List[IOTrial] = [
    IOTrial(
        "fbc2",
        "mini.pickle",
        "mini_fbc2.xml",
        read_sbml_model,
        write_sbml_model,
        validate_sbml_model,
    ),
    IOTrial(
        "fbc2Gz",
        "mini.pickle",
        "mini_fbc2.xml.gz",
        read_sbml_model,
        write_sbml_model,
        None,
    ),
    IOTrial(
        "fbc2Bz2",
        "mini.pickle",
        "mini_fbc2.xml.bz2",
        read_sbml_model,
        write_sbml_model,
        None,
    ),
    IOTrial(
        "fbc1", "mini.pickle", "mini_fbc1.xml", read_sbml_model, write_sbml_model, None
    ),
    IOTrial("cobra", None, "mini_cobra.xml", read_sbml_model, write_sbml_model, None),
]
trial_names: list = [node.name for node in trials]


@pytest.mark.parametrize("trial", trials)
def test_validate(trial: IOTrial, data_directory: Path) -> None:
    """Test validation function.

    Parameters
    ----------
    trial: IOTrial
        Which model trial to check.
    data_directory: Path
        Directory where the data is.
    """
    if trial.validation_function is None:
        pytest.skip("not implemented")
    test_file = data_directory / trial.test_file
    trial.validation_function(test_file)


class TestCobraIO:
    """Tests the read and write functions."""

    @classmethod
    def compare_models(cls, name: str, model1: Model, model2: Model) -> None:
        """Compare two models.

        name, str
            name of models to compare.
        model1: Model
            First model to compare.
        model2: Model
            Second model to compare.
        """
        print(name)
        assert len(model1.reactions) == len(model2.reactions)
        assert len(model1.metabolites) == len(model2.metabolites)
        assert model1.objective.direction == model2.objective.direction
        for attr in (
            "id",
            "name",
            "lower_bound",
            "upper_bound",
            "objective_coefficient",
            "gene_reaction_rule",
        ):
            assert getattr(model1.reactions[0], attr) == getattr(
                model2.reactions[0], attr
            )
            assert getattr(model1.reactions[5], attr) == getattr(
                model2.reactions[5], attr
            )
            assert getattr(model1.reactions[-1], attr) == getattr(
                model2.reactions[-1], attr
            )
        for attr in ("id", "name", "compartment", "formula", "charge"):
            assert getattr(model1.metabolites[0], attr) == getattr(
                model2.metabolites[0], attr
            )
            assert getattr(model1.metabolites[5], attr) == getattr(
                model2.metabolites[5], attr
            )
            assert getattr(model1.metabolites[-1], attr) == getattr(
                model2.metabolites[-1], attr
            )
        assert len(model1.reactions[0].metabolites) == len(
            model2.reactions[0].metabolites
        )
        assert len(model1.reactions[8].metabolites) == len(
            model2.reactions[8].metabolites
        )
        assert len(model1.reactions[-1].metabolites) == len(
            model2.reactions[-1].metabolites
        )
        assert len(model1.genes) == len(model2.genes)

        # ensure they have the same solution max
        solution1 = model1.optimize()
        solution2 = model2.optimize()
        assert (
            solution1.status == "infeasible" and solution2.status == "infeasible"
        ) or abs(solution1.objective_value - solution2.objective_value) < 0.001
        # ensure the references are correct
        assert model2.metabolites[0]._model is model2
        assert model2.reactions[0]._model is model2
        assert model2.genes[0]._model is model2

    @classmethod
    def extra_comparisons(cls, name: str, model1: Model, model2: Model) -> None:
        """Compare additional features of the model.

        name, str
            name of models to compare.
        model1: Model
            First model to compare.
        model2: Model
            Second model to compare.
        """
        print(name)
        assert model1.compartments == model2.compartments

        # FIXME: problems of duplicate annotations in test data
        #  ('cas': ['56-65-5', '56-65-5'])
        # assert dict(model1.metabolites[4].annotation) == dict(
        #    model2.metabolites[4].annotation)
        d1 = model1.reactions[4].annotation
        d2 = model2.reactions[4].annotation
        assert list(d1.keys()) == list(d2.keys())
        for k in d1:
            assert set(d1[k]) == set(d2[k])
        assert dict(model1.reactions[4].annotation) == dict(
            model2.reactions[4].annotation
        )
        assert dict(model1.genes[5].annotation) == dict(model2.genes[5].annotation)

        for attr in ("id", "name"):
            assert getattr(model1.genes[0], attr) == getattr(model2.genes[0], attr)
            assert getattr(model1.genes[10], attr) == getattr(model2.genes[10], attr)
            assert getattr(model1.genes[-1], attr) == getattr(model2.genes[-1], attr)

    def test_read_1(self, io_trial: Tuple[str, Model, Model, Model]) -> None:
        """Read the first model from a processed IOTrial via io_trial().

        io_trial: [str, Model, Model, Model]
            A tuple containing name, reference_model, test_model, rewritten_model.
        """
        name, reference_model, test_model, _ = io_trial
        if name in ["fbc1"]:
            pytest.xfail("not supported")
        if reference_model:
            self.compare_models(name, reference_model, test_model)

    def test_read_2(self, io_trial: Tuple[str, Model, Model, Model]) -> None:
        """Read the second model from a processed IOTrial via io_trial().

        io_trial: [str, Model, Model, Model]
            A tuple containing name, reference_model, test_model, rewritten_model.
        """
        name, reference_model, test_model, _ = io_trial
        if name in ["fbc1", "mat", "cobra", "raven-mat"]:
            pytest.xfail("not supported")
        if reference_model:
            self.extra_comparisons(name, reference_model, test_model)

    def test_write_1(self, io_trial: Tuple[str, Model, Model, Model]) -> None:
        """Test writing the first model from a processed IOTrial via io_trial().

        io_trial: [str, Model, Model, Model]
            A tuple containing name, reference_model, test_model, rewritten_model.
        """
        name, _, test_model, reread_model = io_trial
        if name in ["fbc1", "raven-mat"]:
            pytest.xfail("not supported")

        self.compare_models(name, test_model, reread_model)

    def test_write_2(self, io_trial: Tuple[str, Model, Model, Model]) -> None:
        """Test writing the second model from a processed IOTrial via io_trial().

        io_trial: [str, Model, Model, Model]
            A tuple containing name, reference_model, test_model, rewritten_model.
        """
        name, _, test_model, reread_model = io_trial
        if name in ["fbc1", "mat", "cobra", "raven-mat"]:
            pytest.xfail("not supported")
        self.extra_comparisons(name, test_model, reread_model)


@pytest.fixture(scope="module", params=trials, ids=trial_names)
def io_trial(
    request: SubRequest, data_directory: Path
) -> Tuple[str, Model, Model, Model]:
    """Read reference model, test model, write test model and reread it.

    Parameters
    ----------
    request: IOTrail
    data_directory: Path
        Directory where the data is.

    This function will read the reference model, the test model. It will then write
    the test model based on the write_function() in IOTrial, and read the written test
    model using read_function in IOTrail.
    This can be used to compare the original model, test model and the written and
    reread model.

    Returns
    -------
    Tuple: str, Model, Model, Model
        Name, original model as read by read_function(), test model as read by
        read_function(), test model as written by write_function() and reread by
        read_function().
    """
    reference_model = None
    if request.param.reference_file:
        with open(
            data_directory.joinpath(request.param.reference_file), "rb"
        ) as infile:
            reference_model = load(infile)
    test_model = request.param.read_function(data_directory / request.param.test_file)
    test_output_filename = join(gettempdir(), split(request.param.test_file)[-1])
    # test writing the model within a context with a non-empty stack
    with test_model:
        test_model.objective = test_model.objective
        request.param.write_function(test_model, test_output_filename)
    reread_model = request.param.read_function(test_output_filename)
    unlink(test_output_filename)
    return request.param.name, reference_model, test_model, reread_model


def test_filehandle(data_directory: Path, tmp_path: Path) -> None:
    """Test reading and writing to file handle.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    tmp_path: Path
        Directory to use for temporary data.
    """
    with data_directory.joinpath("mini_fbc2.xml").open("r") as f_in:
        model1 = read_sbml_model(f_in)
        assert model1 is not None

    sbml_path = tmp_path / "test.xml"
    with sbml_path.open("w") as f_out:
        write_sbml_model(model1, f_out)

    with sbml_path.open("r") as f_in:
        model2 = read_sbml_model(f_in)

    TestCobraIO.compare_models(name="filehandle", model1=model1, model2=model2)


def test_from_sbml_string(data_directory: Path) -> None:
    """Test reading from SBML string.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    """
    sbml_path = data_directory / "mini_fbc2.xml"
    model1 = read_sbml_model(sbml_path.read_text())

    model2 = read_sbml_model(sbml_path)
    TestCobraIO.compare_models(name="read from string", model1=model1, model2=model2)


@pytest.mark.skip(reason="Model history currently not written")
def test_model_history(tmp_path: Path) -> None:
    """Testing reading and writing of ModelHistory.

    Parameters
    ----------
    tmp_path: Path
        Directory to use for temporary data.
    """
    model = Model("test")
    model._sbml = {
        "creators": [
            {
                "familyName": "Mustermann",
                "givenName": "Max",
                "organisation": "Muster University",
                "email": "muster@university.com",
            }
        ]
    }

    sbml_path = tmp_path / "test.xml"
    with open(sbml_path, "w") as f_out:
        write_sbml_model(model, f_out)

    with open(sbml_path, "r") as f_in:
        model2 = read_sbml_model(f_in)

    assert "creators" in model2._sbml
    assert len(model2._sbml["creators"]) == 1
    c = model2._sbml["creators"][0]
    assert c["familyName"] == "Mustermann"
    assert c["givenName"] == "Max"
    assert c["organisation"] == "Muster University"
    assert c["email"] == "muster@university.com"


def test_groups(data_directory: Path, tmp_path: Path) -> None:
    """Testing reading and writing of groups.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    tmp_path: Path
        Directory to use for temporary data.
    """
    sbml_path = data_directory / "e_coli_core.xml"
    model = read_sbml_model(sbml_path)
    assert model.groups is not None
    assert len(model.groups) == 10
    g1 = model.groups[0]
    assert len(g1.members) == 6

    temp_path = tmp_path / "test.xml"
    with open(temp_path, "w") as f_out:
        write_sbml_model(model, f_out)

    with open(temp_path, "r") as f_in:
        model2 = read_sbml_model(f_in)

        assert model2.groups is not None
        assert len(model2.groups) == 10
        g1 = model2.groups[0]
        assert len(g1.members) == 6


def test_missing_flux_bounds1(data_directory: Path) -> None:
    """Test missing flux bounds in an incorrect model.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    """
    sbml_path = data_directory / "annotation.xml"
    with open(sbml_path, "r") as f_in:
        # missing flux bounds are set to cobra.configuration.bounds
        # noinspection PyTupleAssignmentBalance
        model, errors = validate_sbml_model(f_in, set_missing_bounds=True)
        r1 = model.reactions.R1
        assert r1.lower_bound == config.lower_bound
        assert r1.upper_bound == config.upper_bound


def test_missing_flux_bounds2(data_directory: Path) -> None:
    """Test missing flux bounds set to [-INF, INF].

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    """
    sbml_path = data_directory / "annotation.xml"
    with open(sbml_path, "r") as f_in:
        # missing flux bounds are set to [-INF, INF]
        # noinspection PyTupleAssignmentBalance
        model, errors = validate_sbml_model(f_in, set_missing_bounds=False)
        r1 = model.reactions.R1
        assert r1.lower_bound == config.lower_bound
        assert r1.upper_bound == config.upper_bound


def test_validate2(data_directory: Path) -> None:
    """Test the validation code.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    """
    sbml_path = data_directory / "mini_fbc2.xml"
    with open(sbml_path, "r") as f_in:
        # noinspection PyTupleAssignmentBalance
        model1, errors = validate_sbml_model(f_in, check_modeling_practice=True)
        assert model1
        assert errors
        assert len(errors["SBML_WARNING"]) == 0


def test_validation_warnings(data_directory: Path) -> None:
    """Test the validation warnings.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    """
    sbml_path = data_directory / "validation.xml"
    with open(sbml_path, "r") as f_in:
        # noinspection PyTupleAssignmentBalance
        model1, errors = validate_sbml_model(f_in, check_modeling_practice=True)
        assert model1
        assert errors
        assert len(errors["COBRA_WARNING"]) == 3
        assert "No objective in listOfObjectives" in errors["COBRA_WARNING"]


def test_infinity_bounds(data_directory: Path, tmp_path: Path) -> None:
    """Test infinity bound example.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    tmp_path: Path
        Directory to use for temporary data.
    """
    sbml_path = data_directory / "fbc_ex1.xml"
    model = read_sbml_model(sbml_path)

    # check that simulation works
    solution = model.optimize()
    assert solution is not None

    # check that values are set
    r = model.reactions.get_by_id("EX_X")
    assert r.lower_bound == -float("Inf")
    assert r.upper_bound == float("Inf")

    temp_path = tmp_path / "test.xml"
    with open(temp_path, "w") as f_out:
        write_sbml_model(model, f_out)

    with open(temp_path, "r") as f_in:
        model2 = read_sbml_model(f_in)
        r = model2.reactions.get_by_id("EX_X")
        assert r.lower_bound == -float("Inf")
        assert r.upper_bound == float("Inf")


def test_boundary_conditions(data_directory: Path) -> None:
    """Test infinity bound example.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    """
    sbml_path1 = data_directory / "fbc_ex1.xml"
    model1 = read_sbml_model(sbml_path1)
    sol1 = model1.optimize()

    # model with species boundaryCondition==True
    sbml_path2 = data_directory / "fbc_ex2.xml"
    model2 = read_sbml_model(sbml_path2)
    sol2 = model2.optimize()

    r = model2.reactions.get_by_id("EX_X")
    assert r.lower_bound == config.lower_bound
    assert r.upper_bound == config.upper_bound

    assert sol1.objective_value == sol2.objective_value


def test_bounds_on_write(data_directory: Path, tmp_path: Path) -> None:
    """Test infinity bound example.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    """
    sbml_path1 = data_directory / "fbc_ex1.xml"
    model1 = read_sbml_model(sbml_path1)

    r_x = model1.reactions.get_by_id("EX_X")
    r_y = model1.reactions.get_by_id("EX_Ac")

    r_x.bounds = (config.lower_bound - 1000, config.upper_bound + 1000)
    assert r_x.lower_bound == config.lower_bound - 1000
    assert r_x.upper_bound == config.upper_bound + 1000

    # Global min/max bounds for other reactions should not change before & after write!
    r_y.bounds = (config.lower_bound, config.upper_bound)
    assert r_y.lower_bound == config.lower_bound
    assert r_y.upper_bound == config.upper_bound

    sbml_path = tmp_path / "test.xml"
    with open(sbml_path, "w") as f_out:
        write_sbml_model(model1, f_out)

    with open(sbml_path, "r") as f_in:
        model2 = read_sbml_model(f_in)

    r2_x = model2.reactions.get_by_id("EX_X")
    r2_y = model2.reactions.get_by_id("EX_Ac")

    assert r2_x.lower_bound == r_x.lower_bound
    assert r2_x.upper_bound == r_x.upper_bound
    assert r2_y.lower_bound == r_y.lower_bound  # before fix #1300, this would fail
    assert r2_y.upper_bound == r_y.upper_bound  # before fix #1300, this would fail


def test_gprs(large_model: Model, tmp_path: Path) -> None:
    """Test that GPRs are written and read correctly.

    Parameters
    ----------
    large_model: Model
        Model to test gprs on.
    tmp_path: Path
        Directory to use for temporary data.
    """
    model1 = large_model
    sbml_path = tmp_path / "test.xml"
    with open(sbml_path, "w") as f_out:
        write_sbml_model(model1, f_out)

    with open(sbml_path, "r") as f_in:
        model2 = read_sbml_model(f_in)

    for r1 in model1.reactions:
        rid = r1.id
        r2 = model2.reactions.get_by_id(rid)
        gpr1 = r1.gene_reaction_rule
        gpr2 = r2.gene_reaction_rule

        assert gpr1 == gpr2


def test_identifiers_annotation() -> None:
    """Test annotation with identifiers."""
    from cobra.io.sbml import _parse_annotation_info

    for uri in [
        "http://identifiers.org/chebi/CHEBI:000123",
        "https://identifiers.org/chebi/CHEBI:000123",
        "http://identifiers.org/CHEBI:000123",
        "https://identifiers.org/CHEBI:000123",
    ]:
        data = _parse_annotation_info(uri)
        assert data
        assert data[0] == "chebi"
        assert data[1] == "CHEBI:000123"

    for uri in [
        "http://identifiers.org/taxonomy/9602",
        "https://identifiers.org/taxonomy/9602",
        "http://identifiers.org/taxonomy:9602",
        "https://identifiers.org/taxonomy:9602",
    ]:
        data = _parse_annotation_info(uri)
        assert data
        assert data[0] == "taxonomy"
        assert data[1] == "9602"

    for uri in [
        "http://identifier.org/taxonomy/9602",
        "https://test.com",
    ]:
        data = _parse_annotation_info(uri)
        assert data is None


def test_smbl_with_notes(data_directory: Path, tmp_path: Path) -> None:
    """Test that NOTES in the RECON 2.2 style are written and read correctly.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    tmp_path: Path
        Directory to use for temporary data.
    """
    sbml_path = join(data_directory, "example_notes.xml")
    model = read_sbml_model(sbml_path)
    assert model.metabolites is not None
    metabolite_notes = {
        "2hb_e": {"CHARGE": "-1", "FORMULA": "C4H7O3", "SMILES": "CCC(O)C(O)=O"},
        "nad_e": {
            "CHARGE": "-1",
            "FORMULA": "C21H26N7O14P2",
            "SMILES": "NC(=O)c1ccc[n+](c1)[C@@H]1O[C@H](COP([O-])("
            "=O)OP([O-])(=O)OC[C@H]2O[C@H]([C@H](O)["
            "C@@H]2O)n2cnc3c(N)ncnc23)[C@@H](O)[C@H]1O",
        },
        "h_e": {"CHARGE": "1", "FORMULA": "H", "SMILES": "[1H+]"},
        "2obut_e": {"CHARGE": "-1", "FORMULA": "C4H5O3", "SMILES": "CCC(=O)C([O-])=O"},
        "nadh_e": {
            "CHARGE": "-2",
            "FORMULA": "C21H27N7O14P2",
            "SMILES": "NC(=O)C1=CN(C=CC1)[C@@H]1O[C@H](COP([O-])("
            "=O)OP([O-])(=O)OC[C@H]2O[C@H]([C@H](O)["
            "C@@H]2O)n2cnc3c(N)ncnc23)[C@@H](O)[C@H]1O",
        },
    }
    metabolite_annotations = {
        "2hb_e": {
            "sbo": "SBO:0000247",
            "inchi": "InChI=1S/C4H8O3/c1-2-3(5)4(6)7/h3,5H,2H2,1H3,(H,6,7)",
            "chebi": "CHEBI:1148",
        },
        "nad_e": {
            "sbo": "SBO:0000247",
            "inchi": "InChI=1S/C21H27N7O14P2/c22-17-12-19("
            "25-7-24-17)28(8-26-12)21-16(32)14(30)11("
            "41-21)6-39-44(36,37)42-43(34,35)38-5-10-13(29)15("
            "31)20(40-10)27-3-1-2-9(4-27)18("
            "23)33/h1-4,7-8,10-11,13-16,20-21,29-32H,5-6H2,"
            "(H5-,22,23,24,25,33,34,35,36,37)/p-1/t10-,"
            "11-,13-,14-,15-,16-,20-,21-/m1/s1",
            "chebi": "CHEBI:57540",
        },
        "h_e": {
            "sbo": "SBO:0000247",
            "inchi": "InChI=1S/p+1/i/hH",
            "chebi": "CHEBI:24636",
        },
        "2obut_e": {
            "sbo": "SBO:0000247",
            "inchi": "InChI=1S/C4H6O3/c1-2-3(5)4(6)7/h2H2,1H3,(H,6,7)/p-1",
            "chebi": "CHEBI:16763",
        },
        "nadh_e": {
            "sbo": "SBO:0000247",
            "inchi": "InChI=1S/C21H29N7O14P2/c22-17-12-19("
            "25-7-24-17)28(8-26-12)21-16(32)14(30)11("
            "41-21)6-39-44(36,37)42-43(34,35)38-5-10-13("
            "29)15(31)20(40-10)27-3-1-2-9(4-27)18("
            "23)33/h1,3-4,7-8,10-11,13-16,20-21,29-32H,2,"
            "5-6H2,(H2,23,33)(H,34,35)(H,36,37)(H2,22,24,"
            "25)/p-2/t10-,11-,13-,14-,15-,16-,20-,21-/m1/s1",
            "chebi": "CHEBI:57945",
        },
    }
    reaction_notes = {
        "CONFIDENCE_LEVEL": "4",
        "NOTES": "NCD",
        "SUBSYSTEM": "Propanoate metabolism",
        "GENE_ASSOCIATION": "(HGNC:8546 and HGNC:8548) or (HGNC:8547 and HGNC:8548)",
    }
    reaction_annotations = {
        "sbo": "SBO:0000176",
        "ec-code": "1.1.1.27",
        "pubmed": ["10108", "21765"],
    }

    for met_id in metabolite_notes:
        assert model.metabolites.has_id(met_id)
        for note_key in metabolite_notes[met_id].keys():
            assert note_key in model.metabolites.get_by_id(met_id).notes
            assert (
                metabolite_notes[met_id][note_key]
                == model.metabolites.get_by_id(met_id).notes[note_key]
            )
        for annotation_key in metabolite_annotations[met_id].keys():
            assert annotation_key in model.metabolites.get_by_id(met_id).annotation
            assert (
                metabolite_annotations[met_id][annotation_key]
                == model.metabolites.get_by_id(met_id).annotation[annotation_key]
            )

    for note_key in reaction_notes.keys():
        assert note_key in model.reactions[0].notes.keys()
        assert reaction_notes[note_key] == model.reactions[0].notes[note_key]
    assert (
        model.reactions[0].gene_reaction_rule == "(HGNC:8546 and HGNC:8548) or "
        "(HGNC:8547 and HGNC:8548)"
    )
    assert len(model.groups) == 1
    for annotation_key in reaction_annotations.keys():
        assert annotation_key in model.reactions[0].annotation.keys()
        assert (
            reaction_annotations[annotation_key]
            == model.reactions[0].annotation[annotation_key]
        )


def test_stable_gprs(data_directory: Path, tmp_path: Path) -> None:
    """Test that GPRs are written correctly after manual changes.

    Parameters
    ----------
    data_directory: Path
        Directory where the data is.
    tmp_path: Path
        Directory to use for temporary data.
    """
    mini = read_sbml_model(join(data_directory, "mini_fbc2.xml"))
    mini.reactions.GLCpts.gene_reaction_rule = "((b2415 and b2417)or (b2416))"
    fixed = join(str(tmp_path), "fixed_gpr.xml")
    write_sbml_model(mini, fixed)
    fixed_model = read_sbml_model(fixed)
    assert (
        fixed_model.reactions.GLCpts.gene_reaction_rule == "(b2415 and b2417) or b2416"
    )


def test_history(data_directory: Path) -> None:
    """Test that the history is read from the model."""
    mini = read_sbml_model(join(data_directory, "mini_history.xml"))
    assert "creators" in mini._sbml
    assert "organisation" in mini._sbml["creators"][0]
    assert "created" in mini._sbml
    assert isinstance(mini._sbml["created"], str)
    assert "2022" in mini._sbml["created"]



"""Define module level fixtures."""

from typing import TYPE_CHECKING

import pytest

from cobra.sampling import ACHRSampler


if TYPE_CHECKING:
    from cobra import Model


@pytest.fixture(scope="function")
def achr(model: "Model") -> ACHRSampler:
    """Return ACHRSampler instance for tests."""
    sampler = ACHRSampler(model, thinning=1)
    assert (sampler.n_warmup > 0) and (sampler.n_warmup <= 2 * len(model.variables))
    assert all(sampler.validate(sampler.warmup) == "v")

    return sampler



"""Test model loading from local and remote model repositores."""

import gzip
import os
from pathlib import Path
from typing import TYPE_CHECKING, Callable
from unittest.mock import Mock

import pytest

from cobra import Configuration
from cobra.io import BiGGModels, BioModels, load_model


if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from cobra import Model


@pytest.fixture(scope="module")
def mini_sbml(cobra_data_directory: Path) -> bytes:
    """Provide a gzip-compressed SBML document."""
    with (cobra_data_directory / "mini_cobra.xml").open(mode="rb") as handle:
        return gzip.compress(handle.read())


@pytest.fixture
def bigg_models(mini_sbml: bytes, mocker: "MockerFixture") -> Mock:
    """Provide a mocked BiGG Models repository interface."""
    result = mocker.Mock(spec_set=BiGGModels)
    result.get_sbml.return_value = mini_sbml
    return result


@pytest.fixture
def biomodels(mini_sbml: bytes, mocker: "MockerFixture") -> Mock:
    """Provide a mocked BioModels repository interface."""
    result = mocker.Mock(spec_set=BioModels)
    result.get_sbml.return_value = mini_sbml
    return result


def test_bigg_access(bigg_models: Mock) -> None:
    """Test that SBML would be retrieved from the BiGG Models repository.

    Parameters
    ----------
    bigg_models : unittest.mock.Mock
        The mocked object for BiGG model respository.

    """
    load_model("e_coli_core", cache=False, repositories=[bigg_models])
    bigg_models.get_sbml.assert_called_once_with(model_id="e_coli_core")


@pytest.mark.xfail(
    condition=os.environ.get("GITHUB_ACTIONS"),
    reason="BioModels often blocks Github Actions",
)
def test_biomodels_access(biomodels: Mock) -> None:
    """Test that SBML would be retrieved from the BioModels repository.

    Parameters
    ----------
    biomodels : unittest.mock.Mock
        The mocked object for BioModels model respository.

    """
    load_model("BIOMD0000000633", cache=False, repositories=[biomodels])
    biomodels.get_sbml.assert_called_once_with(model_id="BIOMD0000000633")


def test_unknown_model() -> None:
    """Expect that a not found error is raised (e2e)."""
    with pytest.raises(RuntimeError):
        load_model("MODELWHO?", cache=False)


@pytest.mark.parametrize(
    "model_id, num_metabolites, num_reactions",
    [("e_coli_core", 72, 95), ("BIOMD0000000633", 50, 35)],
)
def test_remote_load(model_id: str, num_metabolites: int, num_reactions: int) -> None:
    """Test that sample models can be loaded from remote repositories (e2e).

    Parameters
    ----------
    model_id : str
        The ID of the model.
    num_metabolites : int
        The total number of metabolites in the model having ID `model_id`.
    num_reactions : int
        The total number of reactions in the model having ID `model_id`.

    """
    if model_id.startswith("BIOMD"):
        pytest.xfail(reason="BioModels often blocks Github Actions.")

    model = load_model(model_id, cache=False)
    assert len(model.metabolites) == num_metabolites
    assert len(model.reactions) == num_reactions


@pytest.mark.skip("caching currently removed because of a CVE in diskcache")
def test_cache(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path, bigg_models: Mock, biomodels: Mock
) -> None:
    """Test that remote models are properly cached.

    Parameters
    ----------
    monkeypatch : pytest.MonkeyPatch
        The monkeypatch-ing object.
    tmp_path : pathlib.Path
        The path to the temporary test assets store.
    bigg_models : Mock
        The mocked object for BiGG model respository.
    biomodels : unittest.mock.Mock
        The mocked object for BioModels model respository.

    """
    config = Configuration()
    monkeypatch.setattr(config, "cache_directory", tmp_path)
    remote_model = load_model("e_coli_core")
    cached_model = load_model("e_coli_core", repositories=[bigg_models, biomodels])
    bigg_models.get_sbml.assert_not_called()
    biomodels.get_sbml.assert_not_called()
    assert len(cached_model.metabolites) == len(remote_model.metabolites)
    assert len(cached_model.reactions) == len(remote_model.reactions)


def test_local_load(model: "Model", compare_models: Callable) -> None:
    """Test model loading from local repository.

    Parameters
    ----------
    model : cobra.Model
        The model to compare local loading against.
    compare_models : Callable
        A callable object to compare local loading.

    """
    model_local = load_model("textbook")
    compare_models(model, model_local)



"""Provide an implementation of FASTCC."""

from logging import getLogger
from typing import TYPE_CHECKING, List, Optional, Set

from optlang.symbolics import Zero

from .helpers import normalize_cutoff


if TYPE_CHECKING:
    from cobra.core import Model, Reaction

logger = getLogger(__name__)
LARGE_VALUE = 1.0e6


def _add_lp7_vars(
    model: "Model", rxns: List["Reaction"], flux_threshold: float
) -> None:
    """Add the variables and constraints for the LP.

    Parameters
    ----------
    model: cobra.Model
        The model to operate on.
    rxns: list of cobra.Reaction
        The reactions to use for LP.
    flux_threshold: float
        The upper threshold an auxiliary variable can have.

    """
    prob = model.problem
    vars_and_cons = []

    for rxn in rxns:
        var = prob.Variable(f"auxiliary_{rxn.id}", lb=0.0, ub=flux_threshold)
        const = prob.Constraint(
            rxn.flux_expression - var,
            name="aux_constraint_{}".format(rxn.id),
            lb=-LARGE_VALUE,
        )
        vars_and_cons.extend([var, const])

    model.add_cons_vars(vars_and_cons)
    model.solver.update()


def _find_sparse_mode(
    model: "Model", rxn_ids: Set[str], zero_cutoff: float
) -> List["Reaction"]:
    """Perform the LP required for FASTCC.

    Parameters
    ----------
    model: cobra.Model
        The model to perform FASTCC on.
    rxns: list of cobra.Reaction
        The reactions to use for LP.
    zero_cutoff: float
        The cutoff below which flux is considered zero.

    Returns
    -------
    list of cobra.Reaction
        The list of reactions to consider as consistent.

    """
    if not rxn_ids:
        return set()

    # Enable constraints for the reactions
    for rid in rxn_ids:
        model.constraints.get(f"aux_constraint_{rid}").lb = 0.0

    obj_vars = [model.variables.get(f"auxiliary_{rid}") for rid in rxn_ids]
    model.objective = Zero
    model.objective.set_linear_coefficients({v: 1.0 for v in obj_vars})

    sol = model.optimize(objective_sense="max")

    # Disable constraints for the reactions
    for rid in rxn_ids:
        model.constraints.get("aux_constraint_{}".format(rid)).lb = -LARGE_VALUE

    return set(sol.fluxes[sol.fluxes.abs() > zero_cutoff].index)


def _flip_coefficients(model: "Model", rxn_ids: Set[str]) -> None:
    """Flip the coefficients for optimizing in reverse direction.

    Parameters
    ----------
    model: cobra.Model
        The model to operate on.
    rxns: list of cobra.Reaction
        The list of reactions whose coefficients will be flipped.

    """
    if not rxn_ids:
        return
    # flip reactions
    for rxn in rxn_ids:
        const = model.constraints.get(f"aux_constraint_{rxn}")
        var = model.variables.get(f"auxiliary_{rxn}")
        coefs = const.get_linear_coefficients(const.variables)
        const.set_linear_coefficients({k: -v for k, v in coefs.items() if k is not var})
        model.solver.update()


def _any_set(s):
    for x in s:
        return set([x])


def fastcc(
    model: "Model", flux_threshold: float = 1.0, zero_cutoff: Optional[float] = None
) -> "Model":
    r"""
    Check consistency of a metabolic network using FASTCC [1]_.

    FASTCC (Fast Consistency Check) is an algorithm for rapid and
    efficient consistency check in metabolic networks. FASTCC is
    a pure LP implementation and is low on computation resource
    demand. FASTCC also circumvents the problem associated with
    reversible reactions for the purpose. Given a global model,
    it will generate a consistent global model i.e., remove
    blocked reactions. For more details on FASTCC, please
    check [1]_.

    Parameters
    ----------
    model: cobra.Model
        The model to operate on.
    flux_threshold: float, optional
        The flux threshold to consider (default 1.0).
    zero_cutoff: float, optional
        The cutoff to consider for zero flux (default model.tolerance).

    Returns
    -------
    cobra.Model
        The consistent model.

    Notes
    -----
    The LP used for FASTCC is like so:
    maximize: \sum_{i \in J} z_i
    s.t.    : z_i \in [0, \varepsilon] \forall i \in J, z_i \in \mathbb{R}_+
              v_i \ge z_i \forall i \in J
              Sv = 0 v \in B

    References
    ----------
    .. [1] Vlassis N, Pacheco MP, Sauter T (2014)
           Fast Reconstruction of Compact Context-Specific Metabolic Network
           Models.
           PLoS Comput Biol 10(1): e1003424. doi:10.1371/journal.pcbi.1003424

    """
    zero_cutoff = normalize_cutoff(model, zero_cutoff)

    all_rxns = {rxn.id for rxn in model.reactions}
    irreversible_rxns = {rxn.id for rxn in model.reactions if not rxn.reversibility}
    rxns_to_check = irreversible_rxns
    flipped = False
    singletons = False

    with model:
        _add_lp7_vars(model, model.reactions, flux_threshold)

        rxns_to_keep = _find_sparse_mode(model, rxns_to_check, zero_cutoff)
        rxns_to_check = all_rxns.difference(rxns_to_keep)
        logger.info(
            "Initial step found %d consistent reactions. "
            "Starting the consistency loop for the remaining %d reactions.",
            len(rxns_to_keep),
            len(rxns_to_check),
        )

        while rxns_to_check:
            logger.debug(
                "reactions to check: %d - consistent reactions:"
                " %d - flipped: %d - singletons: %d",
                len(rxns_to_check),
                len(rxns_to_keep),
                flipped,
                singletons,
            )
            check = _any_set(rxns_to_check) if singletons else rxns_to_check
            new_rxns = _find_sparse_mode(model, check, zero_cutoff)
            rxns_to_keep.update(new_rxns)

            if rxns_to_check.intersection(rxns_to_keep):
                rxns_to_check = rxns_to_check.difference(rxns_to_keep)
                flipped = False
            else:
                check_irr = check.difference(irreversible_rxns)
                if flipped or not check_irr:
                    if singletons:
                        logger.debug("%s is inconsistent", check)
                        rxns_to_check = rxns_to_check.difference(check)
                    flipped = False
                    singletons = True
                else:
                    flipped = True
                    check = _any_set(rxns_to_check) if singletons else rxns_to_check
                    _flip_coefficients(model, check_irr)
        logger.info(
            "Final - consistent reactions: %d"
            " - inconsistent reactions: %d [eps=%.2g, tol=%.2g]",
            len(rxns_to_keep),
            len(all_rxns) - len(rxns_to_keep),
            flux_threshold,
            zero_cutoff,
        )

    consistent_model = model.copy()
    consistent_model.remove_reactions(
        all_rxns.difference(rxns_to_keep), remove_orphans=True
    )

    return consistent_model



"""Test model annotations in JSON format."""

from pathlib import Path

import pytest

from cobra.io import load_json_model, write_sbml_model


def test_load_json_model_valid(data_directory: Path, tmp_path: Path) -> None:
    """Test loading a valid annotation from JSON.

    data_directory : pathlib.Path
        The path to the test data directory.
    tmp_path : pathlib.Path
        The path to the temporary test assets store.

    """
    path_to_file = data_directory / "valid_annotation_format.json"
    model = load_json_model(path_to_file)
    expected = {
        "bigg.reaction": [["is", "PFK26"]],
        "kegg.reaction": [["is", "R02732"]],
        "rhea": [["is", "15656"]],
    }
    for metabolite in model.metabolites:
        assert metabolite.annotation == expected
    path_to_output = tmp_path / "valid_annotation_output.xml"
    write_sbml_model(model, str(path_to_output.resolve()))


def test_load_json_model_invalid(data_directory: Path) -> None:
    """Test that loading an invalid annotation from JSON raises TypeError.

    data_directory : pathlib.Path
        The path to the test data directory.

    """
    path = data_directory / "invalid_annotation_format.json"
    with pytest.raises(TypeError):
        load_json_model(path)



"""Test functionalities of ROOM."""

from typing import List

import pytest

from cobra import Model
from cobra.flux_analysis.room import add_room


@pytest.mark.parametrize(
    "linear, delta, eps",
    [(True, 0.03, 1e-3), (False, 0.03, 1e-3), (True, 0.1, 1e-2), (False, 0.1, 1e-2)],
)
def test_room_sanity(
    model: Model, all_solvers: List[str], linear: bool, delta: float, eps: float
) -> None:
    """Test optimization criterion and optimality for ROOM."""
    model.solver = all_solvers
    sol = model.optimize()
    with model:
        model.reactions.PYK.knock_out()
        knock_sol = model.optimize()

    with model:
        # let it calculate its own reference solution (pFBA)
        add_room(model, linear=linear, delta=delta, epsilon=eps)
        model.reactions.PYK.knock_out()
        room_sol = model.optimize()

    with model:
        # use the more distant FBA reference
        add_room(model, solution=sol, linear=linear, delta=delta, epsilon=eps)
        model.reactions.PYK.knock_out()
        room_sol_ref = model.optimize()

    # The
    flux_change = (sol.fluxes - knock_sol.fluxes).abs()
    flux_change_room = (sol.fluxes - room_sol.fluxes).abs()
    flux_change_room_ref = (sol.fluxes - room_sol_ref.fluxes).abs()
    rxn_count_naive = (flux_change > delta * sol.fluxes.abs() + eps + 1e-6).sum()
    rxn_count_room = (flux_change_room > delta * sol.fluxes.abs() + eps + 1e-6).sum()
    rxn_count_room_ref = (
        flux_change_room_ref > delta * sol.fluxes.abs() + eps + 1e-6
    ).sum()

    # Expect the ROOM solution to have less changed reactions then a pFBA or FBA
    assert rxn_count_room <= rxn_count_naive
    assert rxn_count_room_ref <= rxn_count_naive



"""Test functionalities of ACHRSampler."""

from typing import TYPE_CHECKING, Callable

import numpy as np
import pytest

from cobra.sampling import ACHRSampler


if TYPE_CHECKING:
    from cobra import Model


def test_achr_init_benchmark(model: "Model", benchmark: Callable) -> None:
    """Benchmark inital ACHR sampling."""
    benchmark(lambda: ACHRSampler(model))


def test_achr_sample_benchmark(achr: ACHRSampler, benchmark: Callable) -> None:
    """Benchmark ACHR sampling."""
    benchmark(achr.sample, 1)


def test_validate_wrong_sample(achr: ACHRSampler, model: "Model") -> None:
    """Test sample correctness."""
    s = achr.sample(10)
    s["hello"] = 1

    with pytest.raises(ValueError):
        achr.validate(s)


def test_sampling(achr: ACHRSampler) -> None:
    """Test sampling."""
    s = achr.sample(10)
    assert all(achr.validate(s) == "v")


def test_batch_sampling(achr: ACHRSampler) -> None:
    """Test batch sampling."""
    for b in achr.batch(5, 4):
        assert all(achr.validate(b) == "v")


def test_variables_samples(achr: ACHRSampler) -> None:
    """Test variable samples."""
    vnames = np.array([v.name for v in achr.model.variables])
    s = achr.sample(10, fluxes=False)
    assert s.shape == (10, achr.warmup.shape[1])
    assert (s.columns == vnames).all()
    assert (achr.validate(s) == "v").all()



"""Unit test the ReactionSummary class."""

import pytest

from cobra.flux_analysis import flux_variability_analysis, pfba
from cobra.summary import ReactionSummary


def test_reaction_summary_interface(model, opt_solver):
    """Test that a summary can be created successfully."""
    model.solver = opt_solver
    reaction = model.reactions.get_by_id("FUM")
    ReactionSummary(
        reaction=reaction,
        model=model,
    )
    ReactionSummary(
        reaction=reaction,
        model=model,
        solution=pfba(model),
    )
    ReactionSummary(
        reaction=reaction,
        model=model,
        fva=0.95,
    )
    ReactionSummary(
        reaction=reaction,
        model=model,
        fva=flux_variability_analysis(model, reaction_list=["FUM"]),
    )


def test_reaction_summary_to_frame(model, opt_solver):
    """Test that the summary's method ``to_frame`` can be called."""
    model.solver = opt_solver
    summary = model.reactions.get_by_id("FUM").summary()
    summary.to_frame()


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"names": True},
        {"float_format": ".1f"},
        {"threshold": 1e-2},
        {"column_width": 20},
    ],
)
def test_reaction_summary_to_string(model, opt_solver, kwargs):
    """Test that the summary's method ``to_string`` can be called."""
    model.solver = opt_solver
    summary = model.reactions.get_by_id("FUM").summary()
    summary.to_string(**kwargs)


@pytest.mark.parametrize(
    "kwargs", [{}, {"names": True}, {"float_format": ".1f"}, {"threshold": 1e-2}]
)
def test_reaction_summary_to_html(model, opt_solver, kwargs):
    """Test that the summary's method ``to_html`` can be called."""
    model.solver = opt_solver
    summary = model.reactions.get_by_id("FUM").summary()
    summary.to_html(**kwargs)


@pytest.mark.parametrize(
    "reaction_id, expected", [("ACALD", 0.0), ("FUM", 5.06), ("PFK", 7.48)]
)
def test_reaction_summary_flux(model, reaction_id: str, expected: float) -> None:
    """Test that the reported flux in the summary is reasonable."""
    result = ReactionSummary(
        reaction=model.reactions.get_by_id(reaction_id), model=model
    )
    assert result.to_frame().at[reaction_id, "flux"] == pytest.approx(
        expected, abs=1e-2
    )


@pytest.mark.parametrize(
    "reaction_id, min_flux, max_flux",
    [("ACALD", -1.27, 0.0), ("FUM", 0.79, 7.38), ("PFK", 2.58, 16.38)],
)
def test_reaction_summary_flux_fva(
    model, reaction_id: str, min_flux: float, max_flux: float
) -> None:
    """Test that the reported flux ranges in the summary are reasonable."""
    result = ReactionSummary(
        reaction=model.reactions.get_by_id(reaction_id), model=model, fva=0.95
    )
    assert result.to_frame().at[reaction_id, "minimum"] == pytest.approx(
        min_flux, abs=1e-2
    )
    assert result.to_frame().at[reaction_id, "maximum"] == pytest.approx(
        max_flux, abs=1e-2
    )


@pytest.mark.parametrize("reaction_id", ["ACALD", "FUM", "PFK"])
def test_reaction_summary_flux_in_context(model, reaction_id: str) -> None:
    """Test that the reaction summary inside and outside of a context are equal."""
    with model:
        context_summary = model.reactions.get_by_id(reaction_id).summary()
    outside_summary = model.reactions.get_by_id(reaction_id).summary()

    assert context_summary.to_frame()["flux"].values == pytest.approx(
        outside_summary.to_frame()["flux"].values, abs=model.tolerance
    )



"""Test functions of array.py."""

from typing import TYPE_CHECKING

import numpy as np
import pytest

from cobra.util import create_stoichiometric_matrix


if TYPE_CHECKING:
    from cobra import Model


def test_dense_matrix(model: "Model") -> None:
    """Test dense stoichiometric matrix creation."""
    S = create_stoichiometric_matrix(model, array_type="dense", dtype=int)
    assert S.dtype == int
    assert np.allclose(S.max(), [59])

    S_df = create_stoichiometric_matrix(model, array_type="DataFrame", dtype=int)
    assert S_df.values.dtype == int
    assert np.all(S_df.columns == [r.id for r in model.reactions])
    assert np.all(S_df.index == [m.id for m in model.metabolites])
    assert np.allclose(S_df.values, S)

    S = create_stoichiometric_matrix(model, array_type="dense", dtype=float)
    solution = model.optimize()
    mass_balance = S.dot(solution.fluxes)
    assert np.allclose(mass_balance, 0)


def test_sparse_matrix(model: "Model") -> None:
    """Test sparse stoichiometric matrix creation."""
    pytest.importorskip("scipy")
    sparse_types = ["dok", "lil"]

    solution = model.optimize()
    for sparse_type in sparse_types:
        S = create_stoichiometric_matrix(model, array_type=sparse_type)
        mass_balance = S.dot(solution.fluxes)
        assert np.allclose(mass_balance, 0)



"""Test finding cyclic reactions in metabolic model."""

from typing import Callable, List

import pytest

from cobra import Model
from cobra.flux_analysis.find_cyclic_reactions import find_cyclic_reactions


@pytest.mark.parametrize("method", ["basic", "optimized"])
def test_find_cyclic_reactions_benchmark(
    large_model: Model, benchmark: Callable, all_solvers: List[str], method: str
) -> None:
    """Benchmark find_cyclic_reactions."""
    large_model.solver = all_solvers
    benchmark(
        find_cyclic_reactions,
        large_model,
        method=method,
    )


@pytest.mark.parametrize("method", ["basic", "optimized"])
def test_find_cyclic_reactions(
    model: Model, all_solvers: List[str], method: str
) -> None:
    """Test find_cyclic_reactions."""
    model.solver = all_solvers
    cyclic_reactions, cyclic_directions = find_cyclic_reactions(model, method=method)

    assert len(cyclic_reactions) == 2
    assert len(cyclic_directions) == 2
    for reaction in ["FRD7", "SUCDi"]:
        assert reaction in cyclic_reactions, f"Expected {reaction} in cyclic reactions"
    for directions in cyclic_directions:
        assert directions == (False, True)



"""Test functionalities of Geometric FBA."""

from typing import Callable, List

import numpy as np
import pandas as pd
import pytest

from cobra.core import Metabolite, Model, Reaction
from cobra.flux_analysis import geometric_fba


@pytest.fixture(scope="module")
def geometric_fba_model() -> Model:
    """Generate geometric FBA model as described in [1]_ .

    References
    ----------
    .. [1] Smallbone, Kieran & Simeonidis, Vangelis. (2009).
           Flux balance analysis: A geometric perspective.
           Journal of theoretical biology.258. 311-5.
           10.1016/j.jtbi.2009.01.027.

    """
    test_model = Model("geometric_fba_paper_model")

    test_model.add_metabolites(Metabolite("A"))
    test_model.add_metabolites(Metabolite("B"))

    v_1 = Reaction("v1", upper_bound=1.0)
    v_1.add_metabolites({test_model.metabolites.A: 1.0})

    v_2 = Reaction("v2", lower_bound=-1000.0)
    v_2.add_metabolites({test_model.metabolites.A: -1.0, test_model.metabolites.B: 1.0})

    v_3 = Reaction("v3", lower_bound=-1000.0)
    v_3.add_metabolites({test_model.metabolites.A: -1.0, test_model.metabolites.B: 1.0})

    v_4 = Reaction("v4", lower_bound=-1000.0)
    v_4.add_metabolites({test_model.metabolites.A: -1.0, test_model.metabolites.B: 1.0})

    v_5 = Reaction("v5")
    v_5.add_metabolites({test_model.metabolites.A: 0.0, test_model.metabolites.B: -1.0})

    test_model.add_reactions([v_1, v_2, v_3, v_4, v_5])

    test_model.objective = "v5"

    return test_model


def test_geometric_fba_benchmark(
    model: Model, benchmark: Callable, all_solvers: List[str]
) -> None:
    """Benchmark geometric_fba."""
    model.solver = all_solvers
    benchmark(geometric_fba, model, processes=1)


def test_geometric_fba(geometric_fba_model: Model, all_solvers: List[str]) -> None:
    """Test geometric_fba."""
    geometric_fba_model.solver = all_solvers
    geometric_fba_sol = geometric_fba(geometric_fba_model, processes=1)
    expected = pd.Series(
        {"v1": 1.0, "v2": 0.33, "v3": 0.33, "v4": 0.33, "v5": 1.0},
        index=["v1", "v2", "v3", "v4", "v5"],
    )
    assert np.allclose(geometric_fba_sol.fluxes, expected, atol=1e-02)



"""Test functions of solver.py."""

import logging
from typing import TYPE_CHECKING, Optional

import numpy as np
import pytest

from cobra.exceptions import OptimizationError
from cobra.util import solver as su


if TYPE_CHECKING:
    from cobra import Model

stable_optlang = ["glpk", "cplex", "gurobi"]
optlang_solvers = [f"optlang-{s}" for s in stable_optlang if s in su.solvers]


def test_solver_list() -> None:
    """Expect that at least the GLPK solver is found."""
    assert len(su.solvers) >= 1
    assert "glpk" in su.solvers


def test_interface_str() -> None:
    """Test the string representation of solver interfaces."""
    assert su.interface_to_str("nonsense") == "nonsense"
    assert su.interface_to_str("optlang.glpk_interface") == "glpk"
    assert su.interface_to_str("optlang-cplex") == "cplex"


def test_solver_name() -> None:
    """Test that the default LP solver name is GLPK."""
    assert su.get_solver_name() == "glpk"


def test_choose_solver(model: "Model") -> Optional[su.SolverNotFound]:
    """Test that solver switching is working."""
    so = su.choose_solver(model, "glpk")
    assert su.interface_to_str(so) == "glpk"

    if any(s in su.solvers for s in su.qp_solvers):
        qp_choice = su.choose_solver(model, qp=True)
        assert su.interface_to_str(qp_choice) in su.qp_solvers
    else:
        with pytest.raises(su.SolverNotFound):
            su.choose_solver(model, qp=True)


def test_linear_reaction_coefficients(model: "Model") -> None:
    """Test that linear coefficients are identifiable in objective."""
    coefficients = su.linear_reaction_coefficients(model)
    assert coefficients == {model.reactions.Biomass_Ecoli_core: 1}


def test_fail_non_linear_reaction_coefficients(model: "Model") -> None:
    """Test failure of non-linear coefficient identification in reaction."""
    model.solver = "optlang-glpk"

    with pytest.raises(ValueError) as error:
        model.objective = model.problem.Objective(
            model.reactions.ATPM.flux_expression**2
        )
        coefficients = su.linear_reaction_coefficients(model)
        assert coefficients == {}

    assert "GLPK only supports linear objectives." in str(error.value)


def test_add_remove(model: "Model") -> None:
    """Test addition and removal of variables and constraints."""
    v = model.variables
    new_var = model.problem.Variable("test_var", lb=-10, ub=-10)
    new_constraint = model.problem.Constraint(
        v.PGK - new_var, name="test_constraint", lb=0
    )

    su.add_cons_vars_to_problem(model, [new_var, new_constraint])
    assert "test_var" in model.variables.keys()
    assert "test_constraint" in model.constraints.keys()

    su.remove_cons_vars_from_problem(model, [new_var, new_constraint])
    assert "test_var" not in model.variables.keys()
    assert "test_constraint" not in model.constraints.keys()


def test_add_remove_in_context(model: "Model") -> None:
    """Test addition and removal of variables and constraints within context."""
    v = model.variables
    new_var = model.problem.Variable("test_var", lb=-10, ub=-10)

    with model:
        su.add_cons_vars_to_problem(model, [new_var])
        su.remove_cons_vars_from_problem(model, [v.PGM])
        assert "test_var" in model.variables.keys()
        assert "PGM" not in model.variables.keys()

    assert "test_var" not in model.variables.keys()
    assert "PGM" in model.variables.keys()


def test_absolute_expression(model: "Model") -> None:
    """Test addition of an absolute expression."""
    v = model.variables
    with model:
        parts = su.add_absolute_expression(model, 2 * v.PGM, name="test", ub=100)
        assert len(parts) == 3
        assert "test" in model.variables.keys()
        assert "abs_pos_test" in model.constraints.keys()
        assert "abs_neg_test" in model.constraints.keys()
    assert "test" not in model.variables.keys()
    assert "abs_pos_test" not in model.constraints.keys()
    assert "abs_neg_test" not in model.constraints.keys()


@pytest.mark.parametrize("solver", optlang_solvers)
def test_fix_objective_as_constraint(solver: str, model: "Model") -> None:
    """Test fixing present objective as a constraint."""
    model.solver = solver
    opt = model.slim_optimize()
    with model as m:
        su.fix_objective_as_constraint(model, 1.0, name="fixed")
        assert (m.constraints.fixed.expression - m.objective.expression).simplify() == 0
        assert m.constraints.fixed.lb == pytest.approx(opt)
    assert "fixed" not in m.constraints
    su.fix_objective_as_constraint(model, name="fixed")
    assert (
        model.constraints.fixed.expression - model.objective.expression
    ).simplify() == 0
    assert m.constraints.fixed.lb == pytest.approx(opt)
    assert "fixed" in model.constraints


@pytest.mark.parametrize("solver", optlang_solvers)
def test_fix_objective_as_constraint_minimize(model: "Model", solver: str) -> None:
    """Test fixing present objective as a constraint but as a minimization."""
    model.solver = solver
    model.reactions.Biomass_Ecoli_core.bounds = (0.1, 0.1)
    minimize_glucose = model.problem.Objective(
        model.reactions.EX_glc__D_e.flux_expression, direction="min"
    )
    su.set_objective(model, minimize_glucose)
    su.fix_objective_as_constraint(model)
    fx_name = f"fixed_objective_{model.objective.name}"
    constr = model.constraints
    # Ensure that a solution exists on non-GLPK solvers.
    model.slim_optimize()
    assert (constr[fx_name].lb, constr[fx_name].ub) == (
        None,
        model.solver.objective.value,
    )


@pytest.mark.parametrize("solver", optlang_solvers)
def test_add_lp_feasibility(model: "Model", solver: str) -> None:
    """Test functionality to ensure LP feasibility."""
    model.solver = solver

    with model:
        with model:
            su.add_lp_feasibility(model)
            assert "s_plus_succoa_c" in model.variables
            assert np.isclose(model.slim_optimize(), 0.0)

        model.reactions.EX_glc__D_e.lower_bound = 1
        assert np.isnan(model.slim_optimize())
        assert "s_plus_succoa_c" not in model.variables

        su.add_lp_feasibility(model)
        assert np.isclose(model.slim_optimize(), 1.3872307692307695)


@pytest.mark.parametrize("solver", optlang_solvers)
def test_add_lexicographic_constraints(model: "Model", solver: str) -> None:
    """Test addition of lexicographic constraints."""
    model.solver = solver

    rxns = ["Biomass_Ecoli_core", "EX_glc__D_e", "EX_o2_e"]

    with model:
        out = su.add_lexicographic_constraints(model, rxns, ["max", "min", "max"])
        print(model.reactions.Biomass_Ecoli_core.bounds)
        assert np.isclose(model.constraints[-3].lb, out[rxns[0]])
        assert np.isclose(model.constraints[-2].ub, out[rxns[1]])
        assert np.isclose(model.constraints[-1].lb, out[rxns[2]])

    with model:
        su.add_lexicographic_constraints(model, rxns, "max")

    with model:
        su.add_lexicographic_constraints(model, rxns)


def test_time_limit(large_model: "Model") -> None:
    """Test time limit while optimizing a model."""
    if su.interface_to_str(large_model.problem) != "glpk":
        pytest.skip("requires GLPK")

    # It is done like this since optlang accepts inputs in seconds
    # whereas GLPK accepts milliseconds
    large_model.solver.configuration._smcp.tm_lim = 1
    with pytest.warns(UserWarning):
        sol = large_model.optimize()
    assert sol.fluxes is not None

    with pytest.raises(OptimizationError):
        sol = large_model.optimize(raise_error=True)


@pytest.mark.parametrize("solver", [s for s in su.solvers if s == "coinor_cbc"])
def test_specialized_solver_warning(solver, caplog):
    """Test the warning for specialized solvers."""
    with caplog.at_level(logging.WARNING):
        su.check_solver(solver)
    assert "are specialized solvers for" in caplog.text



"""Contains module level fixtures and utility functions."""

from pathlib import Path
from pickle import load
from typing import Callable

import pytest
from optlang.interface import OPTIMAL

from cobra import Model


@pytest.fixture(scope="module")
def mini_model(data_directory: Path) -> Model:
    """Fixture for mini model."""
    with (data_directory / "mini.pickle").open(mode="rb") as infile:
        return load(infile)


@pytest.fixture
def compare_models() -> Callable:
    """Compare models as a fixture.

    Returns
    -------
    _compare_models: Callable
        A function that will assert that the models are identical.

    """

    def _compare_models(model_1: Model, model_2: Model) -> None:
        """Compare two models (only for testing purposes).

        Parameters
        ----------
        model_1: Model
            First model to compare.

        model_2: Model
            Second model to compare.

        This function does not return anything, and does a series of asserts on the
        two models. If the models are not identical, the asserts, and the test that
        called them will fail.

        """
        assert len(model_1.reactions) == len(model_2.reactions)
        assert len(model_1.metabolites) == len(model_2.metabolites)
        assert len(model_1.genes) == len(model_2.genes)
        assert model_1.objective.direction == model_2.objective.direction

        # check Reaction attributes
        for attr in (
            "id",
            "name",
            "lower_bound",
            "upper_bound",
            "objective_coefficient",
            "gene_reaction_rule",
        ):
            assert getattr(model_1.reactions[0], attr) == getattr(
                model_2.reactions[0], attr
            )
            assert getattr(model_1.reactions[5], attr) == getattr(
                model_2.reactions[5], attr
            )
            assert getattr(model_1.reactions[-1], attr) == getattr(
                model_2.reactions[-1], attr
            )

        # check Metabolite attributes
        for attr in ("id", "name", "compartment", "formula", "charge"):
            assert getattr(model_1.metabolites[0], attr) == getattr(
                model_2.metabolites[0], attr
            )
            assert getattr(model_1.metabolites[5], attr) == getattr(
                model_2.metabolites[5], attr
            )
            assert getattr(model_1.metabolites[-1], attr) == getattr(
                model_2.metabolites[-1], attr
            )
            assert len(model_1.reactions[0].metabolites) == len(
                model_2.reactions[0].metabolites
            )
        # TODO: either relax gene attribute checking or fix models for testing.
        # check Gene attributes
        # for attr in ("id", "name"):
        #     assert getattr(model_1.genes[0], attr) == getattr(model_2.genes[0],
        #                                                       attr)
        #     assert getattr(model_1.genes[10], attr) == getattr(model_2.genes[10],
        #                                                        attr)
        #     assert getattr(model_1.genes[-1], attr) == getattr(model_2.genes[-1],
        #                                                        attr)

        assert len(model_1.reactions[8].metabolites) == len(
            model_2.reactions[8].metabolites
        )
        assert len(model_1.reactions[-1].metabolites) == len(
            model_2.reactions[-1].metabolites
        )
        assert len(model_1.genes) == len(model_2.genes)

        # ensure they have the same solution max
        solution_1 = model_1.optimize()
        solution_2 = model_2.optimize()
        if solution_1.status == OPTIMAL and solution_2.status == OPTIMAL:
            assert abs(
                solution_1.objective_value - solution_2.objective_value
            ) == pytest.approx(0.0)
        else:
            assert solution_1.status == solution_2.status

        # ensure the references are correct
        # metabolite -> model reference
        assert model_1.metabolites[0]._model is model_1
        assert model_2.metabolites[0]._model is model_2

        # reaction -> model reference
        assert model_1.reactions[0]._model is model_1
        assert model_2.reactions[0]._model is model_2

        # gene -> model reference
        if model_1.genes:
            assert model_1.genes[0]._model is model_1
            assert model_2.genes[0]._model is model_2

        # extra comparisons
        # assert model_1.compartments == model_2.compartments
        # assert dict(model_1.metabolites[4].annotation) == dict(
        #     model_2.metabolites[4].annotation)
        # assert dict(model_1.reactions[4].annotation) == dict(
        #     model_2.reactions[4].annotation)
        # assert dict(model_1.genes[5].annotation) == dict(
        #     model_2.genes[5].annotation)

    return _compare_models



"""Test functions of metabolite.py."""

from typing import TYPE_CHECKING, Tuple

import pytest

from cobra.core import Metabolite


if TYPE_CHECKING:
    from cobra import Model, Solution


def test_metabolite_formula() -> None:
    """Test metabolite formula correctly leads to element dictionary."""
    met = Metabolite("water")
    met.formula = "H2O"
    assert met.elements == {"H": 2, "O": 1}
    assert met.formula_weight == 18.01528


def test_formula_element_setting(model: "Model") -> None:
    """Test that formula and elements set each other when one is set."""
    met = model.metabolites[1]
    orig_formula = str(met.formula)
    orig_elements = dict(met.elements)
    met.formula = ""
    assert met.elements == {}
    met.elements = orig_elements
    assert met.formula == orig_formula


def test_set_id(solved_model: Tuple["Solution", "Model"]) -> None:
    """Test that setting id leads to change in the model metabolite dictlist."""
    solution, model = solved_model
    met = Metabolite("test")
    with pytest.raises(TypeError):
        met.id = 1
    model.add_metabolites([met])
    with pytest.raises(ValueError):
        met.id = "g6p_c"
    met.id = "test2"
    assert "test2" in model.metabolites
    assert "test" not in model.metabolites


def test_remove_from_model(solved_model: Tuple["Solution", "Model"]) -> None:
    """Test that removing a model removes it from relevant model dictlists."""
    solution, model = solved_model
    met = model.metabolites.get_by_id("g6p_c")
    met.remove_from_model()
    assert not (met.id in model.metabolites)
    assert not (met.id in model.constraints)


def test_repr_html_(model: "Model") -> None:
    """Test HTML represenation is correct for a metabolite."""
    assert "<table>" in model.metabolites.h2o_c._repr_html_()



"""Test functions of formula.py ."""

import pytest

from cobra.core.formula import Formula


def test_formula_init() -> None:
    """Test initialization."""
    f = Formula("H2O")
    assert "Formula H2O" in repr(f)
    f = Formula("H2O", name="Water")
    assert f.name == "Water"


@pytest.mark.parametrize(
    ["formula", "weight"], [["H2O", 18.01528], ["C6H12O6", 180.15588], ["NO3", 62.0049]]
)
def test_formula_weight(formula, weight) -> None:
    """Test molecular weight calculation."""
    assert Formula(formula).weight == pytest.approx(weight)


def test_formula_wrong() -> None:
    """Test incorrect formula elements."""
    with pytest.warns(UserWarning):
        w = Formula("NOX").weight
        assert w is None



# -*- coding: utf-8 -*-
#
# cobra documentation build configuration file, created by
# sphinx-quickstart on Wed Jun 13 19:17:34 2012.
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
from os.path import dirname, join

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
SRC_PATH = join(dirname(dirname(__file__)), "src")
sys.path.insert(0, SRC_PATH)

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "autoapi.extension",
    "nbsphinx",
    "sphinx.ext.viewcode",
]
# Document Python Code
autoapi_dirs = [join(SRC_PATH, "cobra")]
autoapi_add_toctree_entry = False

# Enable typehints
autodoc_typehints = "signature"

# Napoleon settings
napoleon_numpy_docstring = True

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "cobra"
copyright = "2016-2022, The cobrapy core team"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
# This import has to be here.
from cobra import __version__ as release  # noqa: E402

version = ".".join(release.split(".")[:2])

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", ".ipynb_checkpoints"]

pygments_style = "sphinx"

# -- Options for HTML output --------------------------------------------------

mathjax_path = (
    "https://cdn.mathjax.org/mathjax/latest/"
    "MathJax.js?config=TeX-AMS-MML_HTMLorMML"
)

# -- Options for LaTeX output --------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    "papersize": "a4paper",
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    "preamble": r"\usepackage{amsmath,amssymb}",
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "cobra.tex",
        "cobra Documentation",
        "The cobrapy core team",
        "manual",
    )
]

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", "cobra", u"cobra Documentation", [u"The cobrapy core team"], 1)
]

# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "cobra",
        "cobra Documentation",
        "The cobrapy core team",
        "cobra",
        "A package for constraints-based modeling of biological networks",
        "Miscellaneous",
    )
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("http://docs.python.org/", None),
    "numpy": ("http://docs.scipy.org/doc/numpy/", None),
    "scipy": ("http://docs.scipy.org/doc/scipy/reference", None),
}
intersphinx_cache_limit = 10  # days to keep the cached inventories

html_theme = "sphinx_book_theme"



"""Test assessing functions in flux_analysis.reaction."""

from typing import List

from cobra.core import Metabolite, Model, Reaction
from cobra.flux_analysis.reaction import assess


def test_assess(model: Model, all_solvers: List[str]) -> None:
    """Test assess functions."""
    with model:
        assert assess(model, model.reactions.GLCpts, solver=all_solvers) is True
        pyr = model.metabolites.pyr_c
        a = Metabolite("a")
        b = Metabolite("b")
        model.add_metabolites([a, b])
        pyr_a2b = Reaction("pyr_a2b")
        pyr_a2b.add_metabolites({pyr: -1, a: -1, b: 1})
        model.add_reactions([pyr_a2b])
        res = assess(model, pyr_a2b, 0.01, solver=all_solvers)
        expected = {
            "precursors": {a: {"required": 0.01, "produced": 0.0}},
            "products": {b: {"required": 0.01, "capacity": 0.0}},
        }
        assert res == expected



"""Test functionalities of Flux Variability Analysis."""

import os
from typing import Callable, List

import numpy as np
import pandas as pd
import pytest

from cobra import Model
from cobra.exceptions import Infeasible
from cobra.flux_analysis.variability import (
    find_blocked_reactions,
    find_essential_genes,
    find_essential_reactions,
    flux_variability_analysis,
)


# FVA
def test_flux_variability_benchmark(
    large_model: Model, benchmark: Callable, all_solvers: List[str]
) -> None:
    """Benchmark FVA."""
    large_model.solver = all_solvers
    benchmark(
        flux_variability_analysis,
        large_model,
        reaction_list=large_model.reactions[1::100],
        processes=1,
    )


def test_pfba_flux_variability(
    model: Model,
    pfba_fva_results: pd.DataFrame,
    fva_results: pd.DataFrame,
    all_solvers: List[str],
) -> None:
    """Test FVA using pFBA."""
    model.solver = all_solvers
    with pytest.warns(UserWarning):
        flux_variability_analysis(
            model,
            pfba_factor=0.1,
            reaction_list=model.reactions[1::3],
            processes=1,
        )
    fva_out = flux_variability_analysis(
        model, pfba_factor=1.1, reaction_list=model.reactions, processes=1
    )
    fva_out.sort_index(inplace=True)
    assert np.allclose(fva_out, pfba_fva_results)
    abs_fva_out = fva_out.dropna().abs()
    abs_fva_results = fva_results.dropna().abs()
    comparison = np.isclose(abs_fva_out, abs_fva_results) | (
        abs_fva_out < abs_fva_results
    )
    assert comparison["minimum"].all()
    assert comparison["maximum"].all()


@pytest.mark.parametrize("loopless", ["fastSNP", "cycleFreeFlux"])
def test_loopless_pfba_fva(model: Model, loopless: str) -> None:
    """Test loopless FVA using pFBA."""
    loop_reactions = [model.reactions.get_by_id(rid) for rid in ("FRD7", "SUCDi")]
    fva_loopless = flux_variability_analysis(
        model,
        pfba_factor=1.1,
        reaction_list=loop_reactions,
        loopless=loopless,
        processes=1,
    )
    assert np.allclose(fva_loopless["maximum"], fva_loopless["minimum"])


def test_flux_variability(
    model: Model, fva_results: pd.DataFrame, all_solvers: List[str]
) -> None:
    """Test FVA."""
    model.solver = all_solvers
    fva_out = flux_variability_analysis(
        model, reaction_list=model.reactions, processes=1
    )
    fva_out.sort_index(inplace=True)
    assert np.allclose(fva_out, fva_results)


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_parallel_flux_variability(
    model: Model, fva_results: pd.DataFrame, all_solvers: List[str]
) -> None:
    """Test parallel FVA."""
    model.solver = all_solvers
    fva_out = flux_variability_analysis(model, processes=2)
    fva_out.sort_index(inplace=True)
    assert np.allclose(fva_out, fva_results)


# Loopless FVA
@pytest.mark.parametrize("loopless", ["fastSNP", "cycleFreeFlux"])
def test_flux_variability_loopless_benchmark(
    model: Model, benchmark: Callable, all_solvers: List[str], loopless: str
) -> None:
    """Benchmark loopless FVA."""
    model.solver = all_solvers
    benchmark(
        flux_variability_analysis,
        model,
        loopless=loopless,
        reaction_list=model.reactions[1::50],
    )


@pytest.mark.parametrize("loopless", ["fastSNP", "cycleFreeFlux"])
def test_flux_variability_loopless(
    model: Model, all_solvers: List[str], loopless: str
) -> None:
    """Test loopless FVA."""
    model.solver = all_solvers
    loop_reactions = [model.reactions.get_by_id(rid) for rid in ("FRD7", "SUCDi")]
    fva_normal = flux_variability_analysis(model, reaction_list=loop_reactions)
    fva_loopless = flux_variability_analysis(
        model,
        reaction_list=loop_reactions,
        loopless=loopless,
    )

    assert not np.allclose(fva_normal["maximum"], fva_normal["minimum"])
    assert np.allclose(fva_loopless["maximum"], fva_loopless["minimum"])


# Internals (essentiality, blocked reactions)
def test_fva_data_frame(model: Model) -> None:
    """Test DataFrame obtained from FVA."""
    df = flux_variability_analysis(model)
    assert np.all([df.columns.values == ["minimum", "maximum"]])


def test_fva_infeasible(model: Model) -> None:
    """Test FVA infeasibility."""
    infeasible_model = model.copy()
    infeasible_model.reactions.get_by_id("EX_glc__D_e").lower_bound = 0
    # ensure that an infeasible model does not run FVA
    with pytest.raises(Infeasible):
        flux_variability_analysis(infeasible_model)


def test_fva_minimization(model: Model) -> None:
    """Test minimization using FVA."""
    model.objective = model.reactions.EX_glc__D_e
    model.objective_direction = "min"
    solution = flux_variability_analysis(model, fraction_of_optimum=0.95)
    assert solution.at["EX_glc__D_e", "minimum"] == -10.0
    assert solution.at["EX_glc__D_e", "maximum"] == -9.5


def test_find_blocked_reactions_solver_none(model: Model) -> None:
    """Test find_blocked_reactions() [no specific solver]."""
    result = find_blocked_reactions(model, model.reactions[40:46])
    assert result == ["FRUpts2"]


def test_essential_genes(model: Model) -> None:
    """Test find_essential_genes()."""
    essential_genes = {
        "b2779",
        "b1779",
        "b0720",
        "b2416",
        "b2926",
        "b1136",
        "b2415",
    }
    observed_essential_genes = {g.id for g in find_essential_genes(model)}
    assert observed_essential_genes == essential_genes


def test_essential_reactions(model: Model) -> None:
    """Test find_blocked_reactions()."""
    essential_reactions = {
        "GLNS",
        "Biomass_Ecoli_core",
        "PIt2r",
        "GAPD",
        "ACONTb",
        "EX_nh4_e",
        "ENO",
        "EX_h_e",
        "EX_glc__D_e",
        "ICDHyr",
        "CS",
        "NH4t",
        "GLCpts",
        "PGM",
        "EX_pi_e",
        "PGK",
        "RPI",
        "ACONTa",
    }
    observed_essential_reactions = {r.id for r in find_essential_reactions(model)}
    assert observed_essential_reactions == essential_reactions


def test_find_blocked_reactions(model: Model, all_solvers: List[str]) -> None:
    """Test find_blocked_reactions()."""
    model.solver = all_solvers
    result = find_blocked_reactions(model, model.reactions[40:46])
    assert result == ["FRUpts2"]

    result = find_blocked_reactions(model, model.reactions[42:48])
    assert set(result) == {"FUMt2_2", "FRUpts2"}

    result = find_blocked_reactions(model, model.reactions[30:50], open_exchanges=True)
    assert result == []



"""Test data storage and recovery using pickle."""

from pathlib import Path
from pickle import dump, load
from typing import Callable

import pytest

from cobra import Model


@pytest.mark.parametrize("load_function", [load])
def test_read_pickle(
    compare_models: Callable,
    data_directory: Path,
    mini_model: Model,
    load_function: Callable,
):
    """Test the reading of model from pickle."""
    if load_function is None:
        pytest.skip()

    with open(data_directory.joinpath("mini.pickle"), "rb") as infile:
        pickle_model = load_function(infile)

    assert compare_models(mini_model, pickle_model) is None


@pytest.mark.parametrize("dump_function", [dump])
def test_write_pickle(tmp_path: Path, mini_model: Model, dump_function: Callable):
    """Test the writing of model to pickle."""
    if dump_function is None:
        pytest.skip()

    output_file = tmp_path.joinpath("mini.pickle")
    with open(str(output_file), "wb") as outfile:
        dump_function(mini_model, outfile)

    assert output_file.exists()



"""Test proper reading of SBML notes."""

from pathlib import Path

from cobra import Metabolite, Model, Reaction
from cobra.io import read_sbml_model, write_sbml_model


def test_notes(tmp_path: Path) -> None:
    """Test if model notes are written in SBML.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the temporary test assets store.

    """
    path_to_file = tmp_path / "model_notes.xml"

    # making a minimal cobra model to test notes
    model = Model("test_notes_model")
    model.notes["Remark"] = "...Model Notes..."
    met = Metabolite("pyr_c", compartment="c")
    model.add_metabolites([met])
    met.notes["Remark"] = "Note with \n newline"
    rxn = Reaction("R_ATPM")
    model.add_reactions([rxn])
    rxn.notes["Remark"] = "What about me?"
    model.objective_direction = "max"
    model.objective = rxn
    write_sbml_model(model, str(path_to_file.resolve()))

    # reading the model back
    model_after_reading = read_sbml_model(str(path_to_file.resolve()))
    met_after_reading = model_after_reading.metabolites.get_by_id("pyr_c")
    reaction_after_reading = model_after_reading.reactions.get_by_id("R_ATPM")

    # checking if notes are written to model
    assert model_after_reading.notes["Remark"] == "...Model Notes..."

    # checking notes for metabolite and reaction
    assert met_after_reading.notes["Remark"] == "Note with \n newline"
    assert reaction_after_reading.notes["Remark"] == "What about me?"
